""" ``plain`` module.
"""

import re

RE_CHOICE_ROUTE = re.compile(
    r"^(?P<p>[\w/]*)" r"\{(?P<n>\w+):\((?P<c>[\w|]+)\)\}" r"(?P<s>[\w/]*)$"
)


def try_build_choice_route(pattern, finishing=True, kwargs=None, name=None):
    """ If the choince route regular expression match the pattern
        than create a ChoiceRoute instance.
    """
    if isinstance(pattern, ChoiceRoute):
        return pattern
    m = RE_CHOICE_ROUTE.match(pattern)
    if m:
        return ChoiceRoute(pattern, finishing, kwargs, name)
    return None


class ChoiceRoute(object):
    """ Route based on choice match, e.g. {locale:(en|ru)}.
    """

    __slots__ = ("kwargs", "name", "exact_matches", "patterns", "path_format")

    def __init__(self, pattern, finishing=True, kwargs=None, name=None):
        kwargs = kwargs and kwargs.copy() or {}
        if name:
            kwargs["route_name"] = name
        self.kwargs = kwargs
        m = RE_CHOICE_ROUTE.match(pattern)
        prefix, self.name, choice, suffix = m.groups()
        choices = choice.split("|")
        self.exact_matches = [
            (prefix + c + suffix, dict(kwargs, **{self.name: c}))
            for c in choices
        ]
        self.patterns = [(p, (len(p), kw)) for p, kw in self.exact_matches]
        self.path_format = prefix + "%s" + suffix

    def match(self, path):
        """ If the ``path`` matches, return the end of
            substring matched and kwargs. Otherwise
            return ``(-1, None)``.
        """
        for pattern, result in self.patterns:
            if path.startswith(pattern):
                return result
        return (-1, None)

    def path(self, values=None):
        """ Build the path for given route.
        """
        if not values or self.name not in values:
            values = self.kwargs
        return self.path_format % values[self.name]



""" ``router`` module.
"""

from warnings import warn

from wheezy.routing.builders import build_route
from wheezy.routing.config import route_builders as default_route_builders
from wheezy.routing.utils import route_name


def url(pattern, handler, kwargs=None, name=None):
    """ Converts parameters to tupple of length four.
        Used for convenience to name parameters and skip
        unused.
    """
    return pattern, handler, kwargs, name


class PathRouter(object):
    """
    """

    __slots__ = (
        "mapping",
        "match_map",
        "path_map",
        "inner_path_map",
        "route_builders",
    )

    def __init__(self, route_builders=None):
        """
        """
        self.route_builders = route_builders or default_route_builders
        # match
        self.match_map = {}
        self.mapping = []
        # path
        self.path_map = {}
        self.inner_path_map = {}

    def add_route(self, pattern, handler, kwargs=None, name=None):
        """ Adds a pattern to route table
        """
        name = name or route_name(handler)
        if name in self.path_map:  # pragma: nocover
            warn("PathRouter: overriding route: %s." % name)
        # build finishing route
        route = build_route(pattern, True, kwargs, name, self.route_builders)
        self.path_map[name] = route.path
        if route.exact_matches:
            for pattern, kwargs in route.exact_matches:
                if pattern in self.match_map:  # pragma: nocover
                    warn("PathRouter: overriding path: %s." % pattern)
                self.match_map[pattern] = (handler, kwargs)
            route.exact_matches = None
        else:
            self.mapping.append((route.match, handler))

    def include(self, pattern, included, kwargs=None):
        """ Includes nested routes below the current.
        """
        # try build intermediate route
        route = build_route(pattern, False, kwargs, None, self.route_builders)
        if not isinstance(included, PathRouter):
            router = PathRouter(self.route_builders)
            router.add_routes(included)
            included = router
        if route.exact_matches:
            for p, kwargs in route.exact_matches:
                for k, v in included.match_map.items():
                    k = p + k
                    if k in self.match_map:  # pragma: nocover
                        warn("PathRouter: overriding path: %s." % k)
                    h, kw = v
                    self.match_map[k] = (h, dict(kwargs, **kw))
            route.exact_matches = None
            included.match_map = {}
            if included.mapping:
                self.mapping.append((route.match, included))
        else:
            self.mapping.append((route.match, included))
        route_path = route.path
        for name, path in included.path_map.items():
            if name in self.inner_path_map:  # pragma: nocover
                warn("PathRouter: overriding route: %s." % name)
            self.inner_path_map[name] = (route_path, path)
        included.path_map = None
        for name, paths in included.inner_path_map.items():
            if name in self.inner_path_map:  # pragma: nocover
                warn("PathRouter: overriding route: %s." % name)
            self.inner_path_map[name] = tuple([route_path] + list(paths))
        included.inner_path_map = None
        # print('include %s => %s / %s' % (pattern, len(self.match_map),
        #                                 len(included.mapping)))

    def add_routes(self, mapping):
        """ Adds routes represented as a list of tuple
            (pattern, handler, kwargs=None, name=None) to route table.
        """
        for m in mapping:
            length = len(m)
            kwargs, name = None, None
            if length == 2:
                pattern, handler = m
            elif length == 3:
                pattern, handler, kwargs = m
            else:
                pattern, handler, kwargs, name = m
            if isinstance(handler, (tuple, list, PathRouter)):
                self.include(pattern, handler, kwargs)
            else:
                self.add_route(pattern, handler, kwargs, name)
        # print('add_routes => %s / %s' % (len(self.match_map),
        #                                 len(self.mapping)))

    def match(self, path):
        """ Tries to find a match for the given path in route table.
            Returns a tupple of (handler, kwargs)
        """
        if path in self.match_map:
            return self.match_map[path]
        for match, handler in self.mapping:
            matched, kwargs = match(path)
            if matched >= 0:
                # TODO: isinstance(handler, PathRouter)
                match = getattr(handler, "match", None)
                if not match:
                    return handler, kwargs
                handler, kwargs_inner = match(path[matched:])
                if handler:
                    if not kwargs:
                        return handler, kwargs_inner
                    if kwargs_inner:
                        kwargs = dict(kwargs, **kwargs_inner)
                    return handler, kwargs
        return None, {}

    def path_for(self, name, **kwargs):
        """ Returns the url for the given route name.
        """
        if name in self.path_map:
            return self.path_map[name](kwargs)
        else:
            return "".join(
                [path(kwargs) for path in self.inner_path_map[name]]
            )



""" ``utils`` module.
"""

import re

RE_STRIP_NAME = re.compile(r"(Handler|Controller|Page|View)$")
RE_CAMELCASE_TO_UNDERSCOPE_1 = re.compile("(.)([A-Z][a-z]+)")
RE_CAMELCASE_TO_UNDERSCOPE_2 = re.compile("([a-z0-9])([A-Z])")


def route_name(handler):
    """ Return a name for the given handler.
        ``handler`` can be an object, class or callable.

        >>> class Login: pass
        >>> route_name(Login)
        'login'
        >>> l = Login()
        >>> route_name(l)
        'login'
    """

    try:
        name = handler.__name__
    except AttributeError:
        name = handler.__class__.__name__

    return camelcase_to_underscore(strip_name(name))


def strip_name(s):
    """ Strips the name per RE_STRIP_NAME regex.

        >>> strip_name('Login')
        'Login'
        >>> strip_name('LoginHandler')
        'Login'
        >>> strip_name('LoginController')
        'Login'
        >>> strip_name('LoginPage')
        'Login'
        >>> strip_name('LoginView')
        'Login'
        >>> strip_name('LoginHandler2')
        'LoginHandler2'
    """
    return RE_STRIP_NAME.sub("", s)


def camelcase_to_underscore(s):
    """ Convert CamelCase to camel_case.

        >>> camelcase_to_underscore('MainPage')
        'main_page'
        >>> camelcase_to_underscore('Login')
        'login'
    """
    s = RE_CAMELCASE_TO_UNDERSCOPE_1.sub(r"\1_\2", s)
    return RE_CAMELCASE_TO_UNDERSCOPE_2.sub(r"\1_\2", s).lower()


def outer_split(expression, sep="()"):
    """ Splits given ``expression`` by outer most separators.

        >>> outer_split('123')
        ['123']
        >>> outer_split('123(45(67)89)123(45)67')
        ['123', '45(67)89', '123', '45', '67']

        If expression is not balanced raises ``ValueError``.

        >>> outer_split('123(') # doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        ValueError: ...
    """
    assert 2 == len(sep)
    start_sep, end_sep = sep
    start_count = end_count = 0
    parts = []
    part = ""
    for token in expression:
        if token == start_sep:
            if start_count == end_count:
                parts.append(part)
                part = ""
                start_count += 1
                continue
            start_count += 1
        elif token == end_sep:
            end_count += 1
            if start_count == end_count:
                parts.append(part)
                part = ""
                continue
        part += token
    if start_count != end_count:
        raise ValueError("Expression is not balanced")
    parts.append(part)
    return parts



""" ``plain`` module.
"""

import re

RE_PLAIN_ROUTE = re.compile(r"^[\w\./-]+$")


def try_build_plain_route(pattern, finishing=True, kwargs=None, name=None):
    """ If the plain route regular expression match the pattern
        than create a PlainRoute instance.
    """
    if isinstance(pattern, PlainRoute):
        return pattern
    if pattern == "" or RE_PLAIN_ROUTE.match(pattern):
        return PlainRoute(pattern, finishing, kwargs, name)
    return None


class PlainRoute(object):
    """ Route based on string equalty operation.
    """

    __slots__ = ("pattern", "kwargs", "matched", "match", "exact_matches")

    def __init__(self, pattern, finishing, kwargs=None, name=None):
        """ Initializes the route by given ``pattern``. If
            ``finishing`` is True than choose ``equals_math``
            strategy
        """
        kwargs = kwargs and kwargs.copy() or {}
        self.pattern = pattern
        self.matched = len(pattern)
        # Choose match strategy
        if finishing:
            if name:
                kwargs["route_name"] = name
            self.match = self.equals_match
        else:
            self.match = self.startswith_match
        self.exact_matches = ((pattern, kwargs),)
        self.kwargs = kwargs

    def equals_match(self, path):
        """ If the ``path`` exactly equals pattern string,
            return end index of substring matched and a copy
            of ``self.kwargs``.
        """
        return (
            path == self.pattern and (self.matched, self.kwargs) or (-1, None)
        )

    def startswith_match(self, path):
        """ If the ``path`` starts with pattern string, return
            the end of substring matched and ``self.kwargs``.
        """
        return (
            path.startswith(self.pattern)
            and (self.matched, self.kwargs)
            or (-1, None)
        )

    def path(self, values=None):
        """ Build the path for given route by simply returning
            the pattern used during initialization.
        """
        return self.pattern



"""
"""

# flake8: noqa: W605

import re

from wheezy.routing.utils import outer_split


def try_build_regex_route(pattern, finishing=True, kwargs=None, name=None):
    """ There is no special tests to match regex selection
        strategy.
    """
    if isinstance(pattern, RegexRoute):
        return pattern
    return RegexRoute(pattern, finishing, kwargs, name)


class RegexRoute(object):
    """ Route based on regular expression matching.
    """

    __slots__ = (
        "match",
        "path",
        "path_value",
        "name",
        "path_format",
        "kwargs",
        "regex",
    )

    exact_matches = None

    def __init__(self, pattern, finishing=True, kwargs=None, name=None):
        pattern = pattern.lstrip("^").rstrip("$")
        # Choose match strategy
        self.path_format, names = parse_pattern(pattern)
        if kwargs:
            self.kwargs = dict.fromkeys(names, "")
            self.kwargs.update(kwargs)
            if finishing:
                self.kwargs["route_name"] = name
            self.match = self.match_with_kwargs
            self.path = self.path_with_kwargs
            self.path_value = self.path_format % self.kwargs
        else:
            if finishing:
                self.name = name
                self.match = self.match_no_kwargs_finishing
            else:
                self.match = self.match_no_kwargs
            self.path = self.path_no_kwargs

        pattern = "^" + pattern
        if finishing:
            pattern = pattern + "$"
        self.regex = re.compile(pattern)

    def match_no_kwargs(self, path):
        """ If the ``path`` match the regex pattern.
        """
        m = self.regex.match(path)
        if m:
            return m.end(), m.groupdict()
        return -1, None

    def match_no_kwargs_finishing(self, path):
        """ If the ``path`` match the regex pattern.
        """
        m = self.regex.match(path)
        if m:
            kwargs = m.groupdict()
            kwargs["route_name"] = self.name
            return m.end(), kwargs
        return -1, None

    def match_with_kwargs(self, path):
        """ If the ``path`` match the regex pattern.
        """
        m = self.regex.match(path)
        if m:
            kwargs = m.groupdict()
            return (m.end(), dict(self.kwargs, **kwargs))
        return -1, None

    def path_with_kwargs(self, values=None):
        """ Build the path for the given route by substituting
            the named places of the regual expression.

            Specialization case: route was initialized with
            default kwargs.
        """
        if values:
            return self.path_format % dict(self.kwargs, **values)
        else:
            return self.path_value

    def path_no_kwargs(self, values):
        """ Build the path for the given route by substituting
            the named places of the regual expression.

            Specialization case: route was initialized with
            no default kwargs.
        """
        return self.path_format % values


RE_SPLIT = re.compile(r"\<(\w+)\>")


def parse_pattern(pattern):
    """ Returns path_format and names.

        >>> parse_pattern(r'abc/(?P<id>[^/]+)')
        ('abc/%(id)s', ['id'])
        >>> parse_pattern(r'abc/(?P<n>[^/]+)/(?P<x>\\\w+)')
        ('abc/%(n)s/%(x)s', ['n', 'x'])
        >>> parse_pattern(r'(?P<locale>(en|ru))/home')
        ('%(locale)s/home', ['locale'])

        >>> from wheezy.routing.curly import convert
        >>> parse_pattern(convert(r'[{locale:(en|ru)}/]home'))
        ('%(locale)s/home', ['locale'])
        >>> parse_pattern(convert(r'item[/{id:i}]'))
        ('item/%(id)s', ['id'])

        >>> p = convert('{controller:w}[/{action:w}[/{id:i}]]')
        >>> parse_pattern(p)
        ('%(controller)s/%(action)s/%(id)s', ['controller', 'action', 'id'])
    """
    pattern = strip_optional(pattern)
    parts = outer_split(pattern, sep="()")
    if len(parts) % 2 == 1 and not parts[-1]:
        parts = parts[:-1]
    names = [RE_SPLIT.split(p)[1] for p in parts[1::2]]
    parts[1::2] = ["%%(%s)s" % p for p in names]
    return "".join(parts), names


def strip_optional(pattern):
    """ Strip optional regex group flag.

        at the beginning

        >>> strip_optional('((?P<locale>(en|ru))/)?home')
        '(?P<locale>(en|ru))/home'

        at the end

        >>> strip_optional('item(/(?P<id>\\\d+))?')
        'item/(?P<id>\\\d+)'

        nested:

        >>> p = '(?P<controller>\\\w+)(/(?P<action>\\\w+)(/(?P<id>\\\d+))?)?'
        >>> strip_optional(p)
        '(?P<controller>\\\w+)/(?P<action>\\\w+)/(?P<id>\\\d+)'
    """
    if ")?" not in pattern:
        return pattern
    parts = outer_split(pattern, sep="()")
    for i in range(2, len(parts), 2):
        part = parts[i]
        if part.startswith("?"):
            parts[i] = part[1:]
            parts[i - 1] = strip_optional(parts[i - 1])
        else:
            parts[i - 1] = "(%s)" % parts[i - 1]
    return "".join(parts)



""" ``curly`` module.
"""

import re

from wheezy.routing.regex import RegexRoute
from wheezy.routing.utils import outer_split

RE_SPLIT = re.compile(r"(?P<n>{[\w:]+.*?})")


def try_build_curly_route(pattern, finishing=True, kwargs=None, name=None):
    """ Convert pattern expression into regex with
        named groups and create regex route.
    """
    if isinstance(pattern, RegexRoute):
        return pattern
    if RE_SPLIT.search(pattern):
        return RegexRoute(convert(pattern), finishing, kwargs, name)
    return None


patterns = {
    # one or more digits
    "i": r"\d+",
    "int": r"\d+",
    "number": r"\d+",
    "digits": r"\d+",
    # one or more word characters
    "w": r"\w+",
    "word": r"\w+",
    # everything until ``/``
    "s": r"[^/]+",
    "segment": r"[^/]+",
    "part": r"[^/]+",
    # any
    "*": r".+",
    "a": r".+",
    "any": r".+",
    "rest": r".+",
}

default_pattern = "s"


def convert(s):
    """ Convert curly expression into regex with
        named groups.
    """
    parts = outer_split(s, sep="[]")
    parts[1::2] = ["(%s)?" % p for p in map(convert, parts[1::2])]
    parts[::2] = map(convert_single, parts[::2])
    return "".join(parts)


def convert_single(s):
    """ Convert curly expression into regex with
        named groups.
    """
    parts = RE_SPLIT.split(s)
    return "".join(map(replace, parts))


def replace(val):
    """ Replace ``{group_name:pattern_name}`` by regex with
        named groups.
    """
    if val.startswith("{") and val.endswith("}"):
        group_name, pattern_name = parse(val[1:-1])
        pattern = patterns.get(pattern_name, pattern_name)
        return "(?P<%s>%s)" % (group_name, pattern)
    return val


def parse(s):
    """ Parse ``s`` according to ``group_name:pattern_name``.

        There is just ``group_name``, return default
        ``pattern_name``.
    """
    if ":" in s:
        return tuple(s.split(":", 1))
    return s, default_pattern



""" ``route`` module.
"""


class Route(object):
    """ Route abstract contract.
    """

    exact_matches = None

    def match(self, path):
        """ if the ``path`` matches, return the end of
            substring matched and kwargs. Otherwise
            return ``(-1, None)``.
        """
        raise NotImplementedError()

    def path(self, values=None):
        """ Build the path for given route.
        """
        raise NotImplementedError()



""" ``builders`` module.
"""


def build_route(pattern, finishing, kwargs, name, route_builders):
    """ Try to find suitable route builder to create a route.
        Raises ``LookupError`` if none found.
    """
    if not finishing:
        assert not name
    for try_build_route in route_builders:
        route = try_build_route(pattern, finishing, kwargs, name)
        if route:
            return route
    else:
        raise LookupError("No matching route factory found")



#!/usr/bin/env python

import os
import re

from setuptools import setup

extra = {}
try:
    from Cython.Build import cythonize

    p = os.path.join("src", "wheezy", "routing")
    extra["ext_modules"] = cythonize(
        [os.path.join(p, "*.py")],
        exclude=os.path.join(p, "__init__.py"),
        nthreads=2,
        quiet=True,
    )
except ImportError:
    pass

README = open(os.path.join(os.path.dirname(__file__), "README.md")).read()
VERSION = (
    re.search(
        r'__version__ = "(.+)"', open("src/wheezy/routing/__init__.py").read(),
    )
    .group(1)
    .strip()
)

setup(
    name="wheezy.routing",
    version=VERSION,
    description="A lightweight path routing library",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/akorn/wheezy.routing",
    author="Andriy Kornatskyy",
    author_email="andriy.kornatskyy@live.com",
    license="MIT",
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.4",
        "Programming Language :: Python :: 2.5",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="routing path url patterns match web mapping",
    packages=["wheezy", "wheezy.routing"],
    package_dir={"": "src"},
    namespace_packages=["wheezy"],
    zip_safe=False,
    platforms="any",
    **extra
)



""" ``config`` module.
"""

from wheezy.routing.choice import try_build_choice_route
from wheezy.routing.curly import (
    default_pattern as curly_default_pattern,
    patterns as curly_patterns,
    try_build_curly_route,
)
from wheezy.routing.plain import try_build_plain_route
from wheezy.routing.regex import try_build_regex_route

assert curly_default_pattern
assert curly_patterns

route_builders = [
    try_build_plain_route,
    try_build_choice_route,
    try_build_curly_route,
    try_build_regex_route,
]
